@@ -11,7 +11,7 @@ use crate::Endpoint;
 
 use super::{
     connection_pool::{ConnId, ConnectionPool, ConnectionRemover},
-    error::{Error, Result},
+    error::{ConnectionError, RecvError, RpcError, SendError, SerializationError},
     wire_msg::WireMsg,
 };
 use bytes::Bytes;
@@ -45,20 +45,25 @@ impl<I: ConnId> Connection<I> {
     }
 
     /// Priority default is 0. Both lower and higher can be passed in.
-    pub(crate) async fn open_bi(&self, priority: i32) -> Result<(SendStream, RecvStream)> {
+    pub(crate) async fn open_bi(
+        &self,
+        priority: i32,
+    ) -> Result<(SendStream, RecvStream), ConnectionError> {
         let (send_stream, recv_stream) = self.handle_error(self.quic_conn.open_bi().await).await?;
         let send_stream = SendStream::new(send_stream);
-        send_stream.set_priority(priority)?;
+        send_stream.set_priority(priority);
         Ok((send_stream, RecvStream::new(recv_stream)))
     }
 
     /// Send message to peer using a uni-directional stream.
     /// Priority default is 0. Both lower and higher can be passed in.
-    pub(crate) async fn send_uni(&self, msg: Bytes, priority: i32) -> Result<()> {
+    pub(crate) async fn send_uni(&self, msg: Bytes, priority: i32) -> Result<(), SendError> {
         let mut send_stream = self.handle_error(self.quic_conn.open_uni().await).await?;
-        send_stream
-            .set_priority(priority)
-            .map_err(|_| Error::UnknownStream)?;
+
+        // quinn returns `UnknownStream` error if the stream does not exist. We ignore it, on the
+        // basis that operations on the stream will fail instead (and the effect of setting priority
+        // or not is only observable if the stream exists).
+        let _ = send_stream.set_priority(priority);
 
         self.handle_error(send_msg(&mut send_stream, msg).await)
             .await?;
@@ -75,6 +80,10 @@ impl<I: ConnId> Connection<I> {
         }
     }
 
+    pub(crate) fn remote_address(&self) -> SocketAddr {
+        self.quic_conn.remote_address()
+    }
+
     async fn handle_error<T, E>(&self, result: Result<T, E>) -> Result<T, E> {
         if result.is_err() {
             self.remover.remove().await
@@ -94,14 +103,18 @@ impl RecvStream {
         Self { quinn_recv_stream }
     }
 
-    /// Read next message from the stream
-    pub async fn next(&mut self) -> Result<Bytes> {
-        match read_bytes(&mut self.quinn_recv_stream).await {
-            Ok(WireMsg::UserMsg(bytes)) => Ok(bytes),
-            Ok(msg) => Err(Error::UnexpectedMessageType(msg)),
-            Err(error) => Err(error),
+    /// Read next user message from the stream
+    pub async fn next(&mut self) -> Result<Bytes, RecvError> {
+        match self.next_wire_msg().await? {
+            WireMsg::UserMsg(bytes) => Ok(bytes),
+            msg => Err(SerializationError::unexpected(msg).into()),
         }
     }
+
+    /// Read next wire msg from the stream
+    pub(crate) async fn next_wire_msg(&mut self) -> Result<WireMsg, RecvError> {
+        read_bytes(&mut self.quinn_recv_stream).await
+    }
 }
 
 impl Debug for RecvStream {
@@ -127,32 +140,25 @@ impl SendStream {
     /// the priority of a stream with pending data may only take effect after that data has been
     /// transmitted. Using many different priority levels per connection may have a negative
     /// impact on performance.
-    pub fn set_priority(&self, priority: i32) -> Result<()> {
-        self.quinn_send_stream
-            .set_priority(priority)
-            .map_err(|_| Error::UnknownStream)?;
-        Ok(())
-    }
-
-    /// Get the priority of the send stream
-    pub fn priority(&self) -> Result<i32> {
-        self.quinn_send_stream
-            .priority()
-            .map_err(|_| Error::UnknownStream)
+    pub fn set_priority(&self, priority: i32) {
+        // quinn returns `UnknownStream` error if the stream does not exist. We ignore it, on the
+        // basis that operations on the stream will fail instead (and the effect of setting priority
+        // or not is only observable if the stream exists).
+        let _ = self.quinn_send_stream.set_priority(priority);
     }
 
     /// Send a message using the stream created by the initiator
-    pub async fn send_user_msg(&mut self, msg: Bytes) -> Result<()> {
+    pub async fn send_user_msg(&mut self, msg: Bytes) -> Result<(), SendError> {
         send_msg(&mut self.quinn_send_stream, msg).await
     }
 
     /// Send a wire message
-    pub async fn send(&mut self, msg: WireMsg) -> Result<()> {
+    pub(crate) async fn send(&mut self, msg: WireMsg) -> Result<(), SendError> {
         msg.write_to_stream(&mut self.quinn_send_stream).await
     }
 
     /// Gracefully finish current stream
-    pub async fn finish(mut self) -> Result<()> {
+    pub async fn finish(mut self) -> Result<(), SendError> {
         self.quinn_send_stream.finish().await?;
         Ok(())
     }
@@ -165,12 +171,12 @@ impl Debug for SendStream {
 }
 
 // Helper to read the message's bytes from the provided stream
-async fn read_bytes(recv: &mut quinn::RecvStream) -> Result<WireMsg> {
+async fn read_bytes(recv: &mut quinn::RecvStream) -> Result<WireMsg, RecvError> {
     WireMsg::read_from_stream(recv).await
 }
 
 // Helper to send bytes to peer using the provided stream.
-async fn send_msg(mut send_stream: &mut quinn::SendStream, msg: Bytes) -> Result<()> {
+async fn send_msg(mut send_stream: &mut quinn::SendStream, msg: Bytes) -> Result<(), SendError> {
     let wire_msg = WireMsg::UserMsg(msg);
     wire_msg.write_to_stream(&mut send_stream).await?;
     Ok(())
@@ -195,8 +201,7 @@ pub(super) fn listen_for_incoming_connections<I: ConnId>(
                         ..
                     }) => {
                         let peer_address = connection.remote_address();
-                        let id = ConnId::generate(&peer_address)
-                            .map_err(|err| Error::ConnectionIdGeneration(err.to_string()))?;
+                        let id = ConnId::generate(&peer_address);
                         let pool_handle =
                             connection_pool.insert(id, peer_address, connection).await;
                         let _ = connection_tx.send(peer_address).await;
@@ -219,7 +224,6 @@ pub(super) fn listen_for_incoming_connections<I: ConnId>(
                 }
             }
         }
-        Ok::<_, Error>(())
     });
 }
 
@@ -254,28 +258,36 @@ pub(super) fn listen_for_incoming_messages<I: ConnId>(
     });
 }
 
+// unify uni- and bi-stream errors
+#[derive(Debug, thiserror::Error)]
+#[error(transparent)]
+enum StreamError {
+    Uni(#[from] RecvError),
+    Bi(#[from] RpcError),
+}
+
 // Read messages sent by peer in an unidirectional stream.
 async fn read_on_uni_streams(
     uni_streams: &mut quinn::IncomingUniStreams,
     peer_addr: SocketAddr,
     message_tx: Sender<(SocketAddr, Bytes)>,
-) -> Result<()> {
+) -> Result<(), StreamError> {
     while let Some(result) = uni_streams.next().await {
         match result {
-            Err(quinn::ConnectionError::ConnectionClosed(close)) => {
+            Err(error @ quinn::ConnectionError::ConnectionClosed(_)) => {
                 trace!("Connection closed by peer {:?}", peer_addr);
-                return Err(Error::ConnectionClosed(close.into()));
+                return Err(StreamError::Uni(error.into()));
             }
-            Err(quinn::ConnectionError::ApplicationClosed(close)) => {
+            Err(error @ quinn::ConnectionError::ApplicationClosed(_)) => {
                 trace!("Connection closed by peer {:?}.", peer_addr);
-                return Err(Error::ConnectionClosed(close.into()));
+                return Err(StreamError::Uni(error.into()));
             }
             Err(err) => {
                 warn!(
                     "Failed to read incoming message on uni-stream for peer {:?} with: {:?}",
                     peer_addr, err
                 );
-                return Err(Error::from(err));
+                return Err(StreamError::Uni(err.into()));
             }
             Ok(mut recv) => loop {
                 match read_bytes(&mut recv).await {
@@ -284,10 +296,6 @@ async fn read_on_uni_streams(
                         let _ = message_tx.send((peer_addr, bytes)).await;
                     }
                     Ok(msg) => warn!("Unexpected message type: {:?}", msg),
-                    Err(Error::StreamRead(quinn::ReadExactError::FinishedEarly)) => {
-                        warn!("Stream read finished early");
-                        break;
-                    }
                     Err(err) => {
                         warn!(
                             "Failed reading from a uni-stream for peer {:?} with: {:?}",
@@ -308,23 +316,23 @@ async fn read_on_bi_streams<I: ConnId>(
     peer_addr: SocketAddr,
     message_tx: Sender<(SocketAddr, Bytes)>,
     endpoint: &Endpoint<I>,
-) -> Result<()> {
+) -> Result<(), StreamError> {
     while let Some(result) = bi_streams.next().await {
         match result {
-            Err(quinn::ConnectionError::ConnectionClosed(close)) => {
+            Err(error @ quinn::ConnectionError::ConnectionClosed(_)) => {
                 trace!("Connection closed by peer {:?}", peer_addr);
-                return Err(Error::ConnectionClosed(close.into()));
+                return Err(StreamError::Bi(error.into()));
             }
-            Err(quinn::ConnectionError::ApplicationClosed(close)) => {
+            Err(error @ quinn::ConnectionError::ApplicationClosed(_)) => {
                 trace!("Connection closed by peer {:?}.", peer_addr);
-                return Err(Error::ConnectionClosed(close.into()));
+                return Err(StreamError::Bi(error.into()));
             }
             Err(err) => {
                 warn!(
                     "Failed to read incoming message on bi-stream for peer {:?} with: {:?}",
                     peer_addr, err
                 );
-                return Err(Error::from(err));
+                return Err(StreamError::Bi(err.into()));
             }
             Ok((mut send, mut recv)) => loop {
                 match read_bytes(&mut recv).await {
@@ -338,7 +346,7 @@ async fn read_on_bi_streams<I: ConnId>(
                                 peer_addr, error
                             );
 
-                            return Err(error);
+                            return Err(StreamError::Bi(error.into()));
                         }
                     }
                     Ok(WireMsg::EndpointVerificationReq(address_sent)) => {
@@ -352,7 +360,7 @@ async fn read_on_bi_streams<I: ConnId>(
                         {
                             warn!("Failed to handle Endpoint verification request for peer {:?} with: {:?}", peer_addr, error);
 
-                            return Err(error);
+                            return Err(StreamError::Bi(error));
                         }
                     }
                     Ok(msg) => {
@@ -361,10 +369,6 @@ async fn read_on_bi_streams<I: ConnId>(
                             peer_addr, msg
                         );
                     }
-                    Err(Error::StreamRead(quinn::ReadExactError::FinishedEarly)) => {
-                        warn!("Stream finished early");
-                        break;
-                    }
                     Err(err) => {
                         warn!(
                             "Failed reading from a bi-stream for peer {:?} with: {:?}",
@@ -383,7 +387,7 @@ async fn read_on_bi_streams<I: ConnId>(
 async fn handle_endpoint_echo_req(
     peer_addr: SocketAddr,
     send_stream: &mut quinn::SendStream,
-) -> Result<()> {
+) -> Result<(), SendError> {
     trace!("Received Echo Request from peer {:?}", peer_addr);
     let message = WireMsg::EndpointEchoResp(peer_addr);
     message.write_to_stream(send_stream).await?;
@@ -396,7 +400,7 @@ async fn handle_endpoint_verification_req<I: ConnId>(
     addr_sent: SocketAddr,
     send_stream: &mut quinn::SendStream,
     endpoint: &Endpoint<I>,
-) -> Result<()> {
+) -> Result<(), RpcError> {
     trace!(
         "Received Endpoint verification request {:?} from {:?}",
         addr_sent,
@@ -429,40 +433,25 @@ async fn handle_endpoint_verification_req<I: ConnId>(
 
 #[cfg(test)]
 mod tests {
-    use crate::api::QuicP2p;
-    use crate::{config::Config, wire_msg::WireMsg, Error};
-    use anyhow::anyhow;
-    use std::net::{IpAddr, Ipv4Addr};
+    use crate::{tests::new_endpoint, wire_msg::WireMsg};
+    use color_eyre::eyre::{eyre, Result};
 
     #[tokio::test(flavor = "multi_thread")]
-    async fn echo_service() -> Result<(), Error> {
-        let qp2p = QuicP2p::<[u8; 32]>::with_config(
-            Some(Config {
-                local_port: None,
-                local_ip: Some(IpAddr::V4(Ipv4Addr::LOCALHOST)),
-                ..Config::default()
-            }),
-            Default::default(),
-            false,
-        )?;
-
+    async fn echo_service() -> Result<()> {
         // Create Endpoint
-        let (peer1, mut peer1_connections, _, _) = qp2p.new_endpoint().await?;
-        let peer1_addr = peer1.socket_addr();
+        let (peer1, mut peer1_connections, _, _, _) = new_endpoint().await?;
+        let peer1_addr = peer1.public_addr();
 
-        let (peer2, _, _, _) = qp2p.new_endpoint().await?;
-        let peer2_addr = peer2.socket_addr();
+        let (peer2, _, _, _, _) = new_endpoint().await?;
+        let peer2_addr = peer2.public_addr();
 
-        peer2.connect_to(&peer1_addr).await?;
+        let _ = peer2.get_or_connect_to(&peer1_addr).await?;
 
         if let Some(connecting_peer) = peer1_connections.next().await {
             assert_eq!(connecting_peer, peer2_addr);
         }
 
-        let connection = peer1
-            .get_connection(&peer2_addr)
-            .await
-            .ok_or(Error::MissingConnection)?;
+        let connection = peer1.get_or_connect_to(&peer2_addr).await?;
         let (mut send_stream, mut recv_stream) = connection.open_bi(0).await?;
         let message = WireMsg::EndpointEchoReq;
         message
@@ -472,7 +461,7 @@ mod tests {
         if let WireMsg::EndpointEchoResp(addr) = message {
             assert_eq!(addr, peer1_addr);
         } else {
-            anyhow!("Unexpected response to EchoService request");
+            eyre!("Unexpected response to EchoService request");
         }
         Ok(())
     }
