@@ -47,9 +47,7 @@
     unused_results
 )]
 
-mod api;
-mod bootstrap_cache;
-mod config;
+pub mod config;
 mod connection_deduplicator;
 mod connection_pool;
 mod connections;
@@ -57,18 +55,20 @@ mod endpoint;
 mod error;
 #[cfg(not(feature = "no-igd"))]
 mod igd;
-mod peer_config;
-#[cfg(test)]
-mod test_utils;
 mod utils;
 mod wire_msg;
 
-pub use api::QuicP2p;
-pub use config::Config;
+pub use config::{Config, ConfigError};
 pub use connection_pool::ConnId;
 pub use connections::{DisconnectionEvents, RecvStream, SendStream};
 pub use endpoint::{Endpoint, IncomingConnections, IncomingMessages};
-pub use error::{Close, ConnectionError, Error, Result};
+#[cfg(not(feature = "no-igd"))]
+pub use error::UpnpError;
+pub use error::{
+    ClientEndpointError, Close, ConnectionError, EndpointError, InternalConfigError, RecvError,
+    RpcError, SendError, SerializationError, StreamError, TransportErrorCode,
+    UnsupportedStreamOperation,
+};
 
 #[cfg(test)]
 mod tests;
