@@ -7,31 +7,28 @@
 // specific language governing permissions and limitations relating to use of the SAFE Network
 // Software.
 
-use super::{hash, new_qp2p, new_qp2p_with_hcc, random_msg};
-use anyhow::{anyhow, Result};
+use super::{hash, local_addr, new_endpoint, random_msg};
+use crate::{Config, Endpoint};
+use color_eyre::eyre::{bail, eyre, Report, Result};
 use futures::{future, stream::FuturesUnordered, StreamExt};
-use std::{
-    collections::{BTreeSet, HashSet},
-    time::Duration,
-};
+use std::{collections::BTreeSet, time::Duration};
 use tokio::time::timeout;
 use tracing::info;
 use tracing_test::traced_test;
 
 #[tokio::test(flavor = "multi_thread")]
 async fn successful_connection() -> Result<()> {
-    let qp2p = new_qp2p()?;
-    let (peer1, mut peer1_incoming_connections, _, _) = qp2p.new_endpoint().await?;
-    let peer1_addr = peer1.socket_addr();
+    let (peer1, mut peer1_incoming_connections, _, _, _) = new_endpoint().await?;
+    let peer1_addr = peer1.public_addr();
 
-    let (peer2, _, _, _) = qp2p.new_endpoint().await?;
+    let (peer2, _, _, _, _) = new_endpoint().await?;
     peer2.connect_to(&peer1_addr).await?;
-    let peer2_addr = peer2.socket_addr();
+    let peer2_addr = peer2.public_addr();
 
     if let Some(connecting_peer) = peer1_incoming_connections.next().await {
         assert_eq!(connecting_peer, peer2_addr);
     } else {
-        anyhow!("No incoming connection");
+        eyre!("No incoming connection");
     }
 
     Ok(())
@@ -39,13 +36,12 @@ async fn successful_connection() -> Result<()> {
 
 #[tokio::test(flavor = "multi_thread")]
 async fn single_message() -> Result<()> {
-    let qp2p = new_qp2p()?;
-    let (peer1, mut peer1_incoming_connections, mut peer1_incoming_messages, _) =
-        qp2p.new_endpoint().await?;
-    let peer1_addr = peer1.socket_addr();
+    let (peer1, mut peer1_incoming_connections, mut peer1_incoming_messages, _, _) =
+        new_endpoint().await?;
+    let peer1_addr = peer1.public_addr();
 
-    let (peer2, _, _, _) = qp2p.new_endpoint().await?;
-    let peer2_addr = peer2.socket_addr();
+    let (peer2, _, _, _, _) = new_endpoint().await?;
+    let peer2_addr = peer2.public_addr();
 
     // Peer 2 connects and sends a message
     peer2.connect_to(&peer1_addr).await?;
@@ -58,7 +54,7 @@ async fn single_message() -> Result<()> {
     if let Some(connecting_peer) = peer1_incoming_connections.next().await {
         assert_eq!(connecting_peer, peer2_addr);
     } else {
-        anyhow!("No incoming connection");
+        eyre!("No incoming connection");
     }
 
     // Peer 2 gets an incoming message
@@ -66,20 +62,19 @@ async fn single_message() -> Result<()> {
         assert_eq!(source, peer2_addr);
         assert_eq!(message, msg_from_peer2);
     } else {
-        anyhow!("No incoming message");
+        eyre!("No incoming message");
     }
     Ok(())
 }
 
 #[tokio::test(flavor = "multi_thread")]
 async fn reuse_outgoing_connection() -> Result<()> {
-    let qp2p = new_qp2p()?;
-    let (alice, _, _, _) = qp2p.new_endpoint().await?;
-    let alice_addr = alice.socket_addr();
+    let (alice, _, _, _, _) = new_endpoint().await?;
+    let alice_addr = alice.public_addr();
 
-    let (bob, mut bob_incoming_connections, mut bob_incoming_messages, _) =
-        qp2p.new_endpoint().await?;
-    let bob_addr = bob.socket_addr();
+    let (bob, mut bob_incoming_connections, mut bob_incoming_messages, _, _) =
+        new_endpoint().await?;
+    let bob_addr = bob.public_addr();
 
     // Connect for the first time and send a message.
     alice.connect_to(&bob_addr).await?;
@@ -90,14 +85,14 @@ async fn reuse_outgoing_connection() -> Result<()> {
     if let Some(connecting_peer) = bob_incoming_connections.next().await {
         assert_eq!(connecting_peer, alice_addr);
     } else {
-        anyhow!("No incoming connection");
+        eyre!("No incoming connection");
     }
 
     if let Some((source, message)) = bob_incoming_messages.next().await {
         assert_eq!(source, alice_addr);
         assert_eq!(message, msg0);
     } else {
-        anyhow!("No incoming message");
+        eyre!("No incoming message");
     }
 
     // Try connecting again and send a message
@@ -109,28 +104,27 @@ async fn reuse_outgoing_connection() -> Result<()> {
     if let Ok(Some(connecting_peer)) =
         timeout(Duration::from_secs(2), bob_incoming_connections.next()).await
     {
-        anyhow!("Unexpected incoming connection from {}", connecting_peer);
+        eyre!("Unexpected incoming connection from {}", connecting_peer);
     }
 
     if let Some((source, message)) = bob_incoming_messages.next().await {
         assert_eq!(source, alice_addr);
         assert_eq!(message, msg1);
     } else {
-        anyhow!("No incoming message");
+        eyre!("No incoming message");
     }
     Ok(())
 }
 
 #[tokio::test(flavor = "multi_thread")]
 async fn reuse_incoming_connection() -> Result<()> {
-    let qp2p = new_qp2p()?;
-    let (alice, mut alice_incoming_connections, mut alice_incoming_messages, _) =
-        qp2p.new_endpoint().await?;
-    let alice_addr = alice.socket_addr();
+    let (alice, mut alice_incoming_connections, mut alice_incoming_messages, _, _) =
+        new_endpoint().await?;
+    let alice_addr = alice.public_addr();
 
-    let (bob, mut bob_incoming_connections, mut bob_incoming_messages, _) =
-        qp2p.new_endpoint().await?;
-    let bob_addr = bob.socket_addr();
+    let (bob, mut bob_incoming_connections, mut bob_incoming_messages, _, _) =
+        new_endpoint().await?;
+    let bob_addr = bob.public_addr();
 
     // Connect for the first time and send a message.
     alice.connect_to(&bob_addr).await?;
@@ -141,14 +135,14 @@ async fn reuse_incoming_connection() -> Result<()> {
     if let Some(connecting_peer) = bob_incoming_connections.next().await {
         assert_eq!(connecting_peer, alice_addr);
     } else {
-        anyhow!("No incoming connection");
+        eyre!("No incoming connection");
     }
 
     if let Some((source, message)) = bob_incoming_messages.next().await {
         assert_eq!(source, alice_addr);
         assert_eq!(message, msg0);
     } else {
-        anyhow!("No incoming message");
+        eyre!("No incoming message");
     }
 
     // Bob tries to connect to alice and sends a message
@@ -161,28 +155,26 @@ async fn reuse_incoming_connection() -> Result<()> {
     if let Ok(Some(connecting_peer)) =
         timeout(Duration::from_secs(2), alice_incoming_connections.next()).await
     {
-        anyhow!("Unexpected incoming connection from {}", connecting_peer);
+        eyre!("Unexpected incoming connection from {}", connecting_peer);
     }
 
     if let Some((source, message)) = alice_incoming_messages.next().await {
         assert_eq!(source, bob_addr);
         assert_eq!(message, msg1);
     } else {
-        anyhow!("No incoming message");
+        eyre!("No incoming message");
     }
     Ok(())
 }
 
 #[tokio::test(flavor = "multi_thread")]
 async fn disconnection() -> Result<()> {
-    let qp2p = new_qp2p()?;
-    let (alice, mut alice_incoming_connections, _, mut alice_disconnections) =
-        qp2p.new_endpoint().await?;
-    let alice_addr = alice.socket_addr();
+    let (alice, mut alice_incoming_connections, _, mut alice_disconnections, _) =
+        new_endpoint().await?;
+    let alice_addr = alice.public_addr();
 
-    let (bob, mut bob_incoming_connections, _, mut bob_disconnections) =
-        qp2p.new_endpoint().await?;
-    let bob_addr = bob.socket_addr();
+    let (bob, mut bob_incoming_connections, _, mut bob_disconnections, _) = new_endpoint().await?;
+    let bob_addr = bob.public_addr();
 
     // Alice connects to Bob who should receive an incoming connection.
     alice.connect_to(&bob_addr).await?;
@@ -190,22 +182,22 @@ async fn disconnection() -> Result<()> {
     if let Some(connecting_peer) = bob_incoming_connections.next().await {
         assert_eq!(connecting_peer, alice_addr);
     } else {
-        anyhow!("No incoming connection");
+        eyre!("No incoming connection");
     }
 
     // After Alice disconnects from Bob both peers should receive the disconnected event.
-    alice.disconnect_from(&bob_addr).await?;
+    alice.disconnect_from(&bob_addr).await;
 
     if let Some(disconnected_peer) = alice_disconnections.next().await {
         assert_eq!(disconnected_peer, bob_addr);
     } else {
-        anyhow!("Missing disconnection event");
+        eyre!("Missing disconnection event");
     }
 
     if let Some(disconnected_peer) = bob_disconnections.next().await {
         assert_eq!(disconnected_peer, alice_addr);
     } else {
-        anyhow!("Missing disconnection event");
+        eyre!("Missing disconnection event");
     }
 
     // This time bob connects to Alice. Since this is a *new connection*, Alice should get the connection event
@@ -214,7 +206,7 @@ async fn disconnection() -> Result<()> {
     if let Some(connected_peer) = alice_incoming_connections.next().await {
         assert_eq!(connected_peer, bob_addr);
     } else {
-        anyhow!("Missing incoming connection");
+        eyre!("Missing incoming connection");
     }
 
     Ok(())
@@ -226,31 +218,31 @@ async fn simultaneous_incoming_and_outgoing_connections() -> Result<()> {
     // others connection first), two separate connections are created. This test verifies that
     // everything still works correctly even in this case.
 
-    let qp2p = new_qp2p()?;
     let (
         alice,
         mut alice_incoming_connections,
         mut alice_incoming_messages,
         mut alice_disconnections,
-    ) = qp2p.new_endpoint().await?;
-    let alice_addr = alice.socket_addr();
+        _,
+    ) = new_endpoint().await?;
+    let alice_addr = alice.public_addr();
 
-    let (bob, mut bob_incoming_connections, mut bob_incoming_messages, _) =
-        qp2p.new_endpoint().await?;
-    let bob_addr = bob.socket_addr();
+    let (bob, mut bob_incoming_connections, mut bob_incoming_messages, _, _) =
+        new_endpoint().await?;
+    let bob_addr = bob.public_addr();
 
     future::try_join(alice.connect_to(&bob_addr), bob.connect_to(&alice_addr)).await?;
 
     if let Some(connecting_peer) = alice_incoming_connections.next().await {
         assert_eq!(connecting_peer, bob_addr);
     } else {
-        anyhow!("No incoming connection from Bob");
+        eyre!("No incoming connection from Bob");
     }
 
     if let Some(connecting_peer) = bob_incoming_connections.next().await {
         assert_eq!(connecting_peer, alice_addr);
     } else {
-        anyhow!("No incoming connectino from Alice");
+        eyre!("No incoming connectino from Alice");
     }
 
     let msg0 = random_msg(1024);
@@ -263,24 +255,24 @@ async fn simultaneous_incoming_and_outgoing_connections() -> Result<()> {
         assert_eq!(src, bob_addr);
         assert_eq!(message, msg1);
     } else {
-        anyhow!("Missing incoming message");
+        eyre!("Missing incoming message");
     }
 
     if let Some((src, message)) = bob_incoming_messages.next().await {
         assert_eq!(src, alice_addr);
         assert_eq!(message, msg0);
     } else {
-        anyhow!("Missing incoming message");
+        eyre!("Missing incoming message");
     }
 
     // Drop the connection initiated by Bob.
-    bob.disconnect_from(&alice_addr).await?;
+    bob.disconnect_from(&alice_addr).await;
 
     // It should be closed on Alice's side too.
     if let Some(disconnected_peer) = alice_disconnections.next().await {
         assert_eq!(disconnected_peer, bob_addr);
     } else {
-        anyhow!("Missing disconnection event");
+        eyre!("Missing disconnection event");
     }
 
     // Bob connects to Alice again. This does not open a new connection but returns the connection
@@ -290,7 +282,7 @@ async fn simultaneous_incoming_and_outgoing_connections() -> Result<()> {
     if let Ok(Some(connecting_peer)) =
         timeout(Duration::from_secs(2), alice_incoming_connections.next()).await
     {
-        anyhow!("Unexpected incoming connection from {}", connecting_peer);
+        eyre!("Unexpected incoming connection from {}", connecting_peer);
     }
 
     let msg2 = random_msg(1024);
@@ -306,13 +298,12 @@ async fn simultaneous_incoming_and_outgoing_connections() -> Result<()> {
 
 #[tokio::test(flavor = "multi_thread")]
 async fn multiple_concurrent_connects_to_the_same_peer() -> Result<()> {
-    let qp2p = new_qp2p()?;
-    let (alice, mut alice_incoming_connections, mut alice_incoming_messages, _) =
-        qp2p.new_endpoint().await?;
-    let alice_addr = alice.socket_addr();
+    let (alice, mut alice_incoming_connections, mut alice_incoming_messages, _, _) =
+        new_endpoint().await?;
+    let alice_addr = alice.public_addr();
 
-    let (bob, _, mut bob_incoming_messages, _) = qp2p.new_endpoint().await?;
-    let bob_addr = bob.socket_addr();
+    let (bob, _, mut bob_incoming_messages, _, _) = new_endpoint().await?;
+    let bob_addr = bob.public_addr();
 
     // Try to establish two connections to the same peer at the same time.
     let ((), ()) =
@@ -322,13 +313,13 @@ async fn multiple_concurrent_connects_to_the_same_peer() -> Result<()> {
     if let Some(connecting_peer) = alice_incoming_connections.next().await {
         assert_eq!(connecting_peer, bob_addr);
     } else {
-        anyhow!("Missing incoming connection");
+        eyre!("Missing incoming connection");
     }
 
     if let Ok(Some(connecting_peer)) =
         timeout(Duration::from_secs(2), alice_incoming_connections.next()).await
     {
-        anyhow!("Unexpected incoming connection from {}", connecting_peer);
+        eyre!("Unexpected incoming connection from {}", connecting_peer);
     }
 
     // Send two messages, one from each end
@@ -343,14 +334,14 @@ async fn multiple_concurrent_connects_to_the_same_peer() -> Result<()> {
         assert_eq!(src, bob_addr);
         assert_eq!(message, msg1);
     } else {
-        anyhow!("No message received from Bob");
+        eyre!("No message received from Bob");
     }
 
     if let Some((src, message)) = bob_incoming_messages.next().await {
         assert_eq!(src, alice_addr);
         assert_eq!(message, msg0);
     } else {
-        anyhow!("No message from alice");
+        eyre!("No message from alice");
     }
 
     Ok(())
@@ -364,9 +355,8 @@ async fn multiple_connections_with_many_concurrent_messages() -> Result<()> {
     let num_messages_each: usize = 100;
     let num_messages_total: usize = 1000;
 
-    let qp2p = new_qp2p()?;
-    let (server_endpoint, _, mut recv_incoming_messages, _) = qp2p.new_endpoint().await?;
-    let server_addr = server_endpoint.socket_addr();
+    let (server_endpoint, _, mut recv_incoming_messages, _, _) = new_endpoint().await?;
+    let server_addr = server_endpoint.public_addr();
 
     let test_msgs: Vec<_> = (0..num_messages_each).map(|_| random_msg(1024)).collect();
     let sending_msgs = test_msgs.clone();
@@ -398,7 +388,7 @@ async fn multiple_connections_with_many_concurrent_messages() -> Result<()> {
                             .send_message(hash_result.to_vec().into(), &src, 0)
                             .await?;
 
-                        Ok::<_, anyhow::Error>(())
+                        Ok::<_, Report>(())
                     }
                 }));
 
@@ -418,8 +408,7 @@ async fn multiple_connections_with_many_concurrent_messages() -> Result<()> {
     for id in 0..num_senders {
         let messages = sending_msgs.clone();
         tasks.push(tokio::spawn({
-            let qp2p = new_qp2p()?;
-            let (send_endpoint, _, mut recv_incoming_messages, _) = qp2p.new_endpoint().await?;
+            let (send_endpoint, _, mut recv_incoming_messages, _, _) = new_endpoint().await?;
 
             async move {
                 let mut hash_results = BTreeSet::new();
@@ -454,7 +443,7 @@ async fn multiple_connections_with_many_concurrent_messages() -> Result<()> {
                     }
                 }
 
-                Ok::<_, anyhow::Error>(())
+                Ok::<_, Report>(())
             }
         }));
     }
@@ -470,9 +459,8 @@ async fn multiple_connections_with_many_larger_concurrent_messages() -> Result<(
     let num_messages_each: usize = 100;
     let num_messages_total: usize = num_senders * num_messages_each;
 
-    let qp2p = new_qp2p()?;
-    let (server_endpoint, _, mut recv_incoming_messages, _) = qp2p.new_endpoint().await?;
-    let server_addr = server_endpoint.socket_addr();
+    let (server_endpoint, _, mut recv_incoming_messages, _, _) = new_endpoint().await?;
+    let server_addr = server_endpoint.public_addr();
 
     let test_msgs: Vec<_> = (0..num_messages_each)
         .map(|_| random_msg(1024 * 1024))
@@ -526,8 +514,7 @@ async fn multiple_connections_with_many_larger_concurrent_messages() -> Result<(
         assert!(!logs_contain("error"));
 
         tasks.push(tokio::spawn({
-            let qp2p = new_qp2p()?;
-            let (send_endpoint, _, mut recv_incoming_messages, _) = qp2p.new_endpoint().await?;
+            let (send_endpoint, _, mut recv_incoming_messages, _, _) = new_endpoint().await?;
 
             async move {
                 let mut hash_results = BTreeSet::new();
@@ -564,7 +551,7 @@ async fn multiple_connections_with_many_larger_concurrent_messages() -> Result<(
                     }
                 }
 
-                Ok::<_, anyhow::Error>(())
+                Ok::<_, Report>(())
             }
         }));
     }
@@ -572,7 +559,7 @@ async fn multiple_connections_with_many_larger_concurrent_messages() -> Result<(
     while let Some(result) = tasks.next().await {
         match result {
             Ok(Ok(())) => (),
-            other => anyhow::bail!("Error from test threads: {:?}", other??),
+            other => bail!("Error from test threads: {:?}", other??),
         }
     }
     Ok(())
@@ -585,12 +572,11 @@ async fn many_messages() -> Result<()> {
 
     let num_messages: usize = 10_000;
 
-    let qp2p = new_qp2p()?;
-    let (send_endpoint, _, _, _) = qp2p.new_endpoint().await?;
-    let (recv_endpoint, _, mut recv_incoming_messages, _) = qp2p.new_endpoint().await?;
+    let (send_endpoint, _, _, _, _) = new_endpoint().await?;
+    let (recv_endpoint, _, mut recv_incoming_messages, _, _) = new_endpoint().await?;
 
-    let send_addr = send_endpoint.socket_addr();
-    let recv_addr = recv_endpoint.socket_addr();
+    let send_addr = send_endpoint.public_addr();
+    let recv_addr = recv_endpoint.public_addr();
 
     let mut tasks = Vec::new();
 
@@ -607,7 +593,7 @@ async fn many_messages() -> Result<()> {
                 endpoint.send_message(msg, &recv_addr, 0).await?;
                 info!("sent {}", id);
 
-                Ok::<_, anyhow::Error>(())
+                Ok::<_, Report>(())
             }
         }));
     }
@@ -642,19 +628,23 @@ async fn many_messages() -> Result<()> {
 // bootstrap contacts later.
 #[tokio::test(flavor = "multi_thread")]
 async fn connection_attempts_to_bootstrap_contacts_should_succeed() -> Result<()> {
-    let qp2p = new_qp2p()?;
+    let (ep1, _, _, _, _) = new_endpoint().await?;
+    let (ep2, _, _, _, _) = new_endpoint().await?;
+    let (ep3, _, _, _, _) = new_endpoint().await?;
 
-    let (ep1, _, _, _) = qp2p.new_endpoint().await?;
-    let (ep2, _, _, _) = qp2p.new_endpoint().await?;
-    let (ep3, _, _, _) = qp2p.new_endpoint().await?;
+    let contacts = vec![ep1.public_addr(), ep2.public_addr(), ep3.public_addr()];
 
-    let contacts = vec![ep1.socket_addr(), ep2.socket_addr(), ep3.socket_addr()]
-        .iter()
-        .cloned()
-        .collect::<HashSet<_>>();
-
-    let qp2p = new_qp2p_with_hcc(contacts.clone())?;
-    let (ep, _, _, _, bootstrapped_peer) = qp2p.bootstrap().await?;
+    let (ep, _, _, _, bootstrapped_peer) = Endpoint::<[u8; 32]>::new(
+        local_addr(),
+        &contacts,
+        Config {
+            min_retry_duration: Duration::from_millis(500).into(),
+            ..Config::default()
+        },
+    )
+    .await?;
+    let bootstrapped_peer =
+        bootstrapped_peer.ok_or_else(|| eyre!("Failed to connecto to any contact"))?;
 
     for peer in contacts {
         if peer != bootstrapped_peer {
@@ -666,15 +656,41 @@ async fn connection_attempts_to_bootstrap_contacts_should_succeed() -> Result<()
 
 #[tokio::test(flavor = "multi_thread")]
 async fn reachability() -> Result<()> {
-    let qp2p = new_qp2p()?;
-
-    let (ep1, _, _, _) = qp2p.new_endpoint().await?;
-    let (ep2, _, _, _) = qp2p.new_endpoint().await?;
+    let (ep1, _, _, _, _) = new_endpoint().await?;
+    let (ep2, _, _, _, _) = new_endpoint().await?;
 
     if let Ok(()) = ep1.is_reachable(&"127.0.0.1:12345".parse()?).await {
-        anyhow!("Unexpected success");
+        eyre!("Unexpected success");
     };
-    let reachable_addr = ep2.socket_addr();
+    let reachable_addr = ep2.public_addr();
     ep1.is_reachable(&reachable_addr).await?;
     Ok(())
 }
+
+#[tokio::test(flavor = "multi_thread")]
+#[traced_test]
+async fn client() -> Result<()> {
+    use crate::{Config, Endpoint};
+
+    let (server, _, mut server_messages, _, _) = new_endpoint().await?;
+    let client = Endpoint::<[u8; 32]>::new_client(
+        local_addr(),
+        Config {
+            min_retry_duration: Some(Duration::from_millis(500)),
+            ..Default::default()
+        },
+    )?;
+
+    client
+        .send_message(b"hello"[..].into(), &server.public_addr(), 0)
+        .await?;
+
+    let (sender, message) = server_messages
+        .next()
+        .await
+        .ok_or_else(|| eyre!("Did not receive expected message"))?;
+    assert_eq!(sender, client.public_addr());
+    assert_eq!(&message[..], b"hello");
+
+    Ok(())
+}
